# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from pydantic import Field

from psdm.base import Base


class ElementState(Base):
    """This class represents the state of an element which is (partly) disconnected or out of service."""

    name: str
    disabled: bool = False
    open_switches: tuple[str, ...] = Field(default_factory=tuple)



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :author: Sebastian Krahmer <sebastian.krahmer@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from collections.abc import Sequence

import pydantic

from psdm.base import Base

variable_dimension = 2


def validate_dimension(
    value: Sequence[Sequence[float]] | None,
) -> Sequence[Sequence[float]] | None:
    if value is not None and len(value[0]) is not len(value[1]):
        raise ValueError
    if value is not None and len(value) > variable_dimension:
        raise ValueError

    return value


class Characteristic(Base):
    """This class represents a data point based characteristic of power injection for a load."""

    name: str
    description: str | None = None
    data: Sequence[Sequence[float]] | None = None  # intended to be a sequence of two lists, e.g. Q(P)-characteristic

    @pydantic.field_validator("data", mode="before")
    def validate_characteristic_dimension(
        cls,
        v: Sequence[Sequence[float]] | None,
    ) -> Sequence[Sequence[float]] | None:
        return validate_dimension(v)



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

import datetime
import enum
import uuid

import pydantic

from psdm.base import Base

VERSION = "1.8.1"


class SignConvention(enum.Enum):
    PASSIVE = "PASSIVE"  # consumer load centered
    ACTIVE = "ACTIVE"  # producer load centered


class Meta(Base):
    """This class represents the meta data related to the grid export."""

    name: str
    date: datetime.date
    id: uuid.UUID = pydantic.Field(default_factory=uuid.uuid4)  # noqa: A003
    sign_convention: SignConvention | None = None
    project: str | None = None
    case: str | None = None

    @pydantic.computed_field  # type: ignore[misc]
    @property
    def version(self) -> str:
        return VERSION



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :author: Sebastian Krahmer <sebastian.krahmer@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from psdm.base import Base
from psdm.steadystate_case.controller import PController


class ActivePower(Base):
    """This class represents the three phase active power operating point of a load.

    It must be characterized by a controller.
    """

    controller: PController



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from psdm.base import Base
from psdm.base import UniqueTuple
from psdm.meta import Meta
from psdm.topology_case.element_state import ElementState


class Case(Base):
    """This class represents a unique topology case of the grid.

    It is characterized by a list of element which are out of service.
    """

    meta: Meta
    elements: UniqueTuple[ElementState]



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :author: Sebastian Krahmer <sebastian.krahmer@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

import pydantic

from psdm.base import Base


class LoadModel(Base):
    """Load representation based on polynomial model.

    load = load_0*(c_p*(U/U_0)^exp_p + c_i*(U/U_0)^exp_i + (c_z)*(U/U_0)^exp_z)
    c_z = 1 - c_p - c_i
    """

    name: str | None = None
    c_p: pydantic.confloat(ge=0, le=1) = 1.0  # type: ignore[valid-type]
    c_i: pydantic.confloat(ge=0, le=1) = 0.0  # type: ignore[valid-type]
    exp_p: int = 0
    exp_i: int = 1
    exp_z: int = 2

    @pydantic.model_validator(mode="after")  # type: ignore[arg-type]
    def validate_range_c(cls, load_model: LoadModel) -> LoadModel:
        name = load_model.name
        c_p = load_model.c_p
        c_i = load_model.c_i
        if c_p + c_i > 1:
            msg = f"Load model {name!r}: Sum of components must not exceed 1, but {(c_p + c_i)=}."
            raise ValueError(msg)

        return load_model

    @pydantic.computed_field  # type: ignore[misc]
    @property
    def c_z(self) -> float:
        return 1 - self.c_p - self.c_i


CONSTANT_POWER_LM = LoadModel(name="constant_power")



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from psdm.base import Base
from psdm.base import UniqueTuple
from psdm.meta import Meta
from psdm.topology.branch import Branch
from psdm.topology.external_grid import ExternalGrid
from psdm.topology.load import Load
from psdm.topology.node import Node
from psdm.topology.transformer import Transformer


class Topology(Base):
    """This class represents operating point independent topology of a grid.

    It is characterized by list of branches, nodes, loads, transformers and external grids.
    """

    meta: Meta
    branches: UniqueTuple[Branch]
    nodes: UniqueTuple[Node]
    loads: UniqueTuple[Load]
    transformers: UniqueTuple[Transformer]
    external_grids: UniqueTuple[ExternalGrid]



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from contextlib import nullcontext as does_not_raise

import pydantic
import pytest
from psdm.steadystate_case.controller import QController
from psdm.steadystate_case.controller import ControlQConst

from psdm.steadystate_case.reactive_power import ReactivePower


class TestReactivePower:
    @pytest.mark.parametrize(
        (
            "controller",
            "expectation",
        ),
        [
            (
                QController(
                    node_target="Node_A",
                    control_type=ControlQConst(
                        value=0,
                        value_a=0,
                        value_b=0,
                        value_c=0,
                        is_symmetrical=True,
                    ),
                ),
                does_not_raise(),
            ),
            (None, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(  # noqa: PLR0913
        self,
        controller,
        expectation,
    ) -> None:
        with expectation:
            ReactivePower(
                controller=controller,
            )



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

import enum

from psdm.base import Base


class VectorGroup(enum.Enum):
    Y = "Y"
    YN = "YN"
    Z = "Z"
    ZN = "ZN"
    D = "D"


class Winding(Base):
    """This class represents a winding of a transformer.

    For example, a 2-winding transformer has a high and low voltage level winding.
    Each windings is characterized by vector group, which defines the interconnection of the three phases.
    """

    node: str
    s_r: float
    u_n: float  # Nominal Voltage of connected nodes (CIM: BaseVoltage)
    u_r: float  # Rated Voltage of the transformer windings itself (CIM: ratedU)
    r1: float  # positive sequence values of transformer T-representation
    x1: float
    r0: float | None = None  # zero sequence values of transformer T-representation
    x0: float | None = None
    re: float | None = None  # earthing of neutral point
    xe: float | None = None  # earthing of neutral point
    phase_angle_clock: int | None = None
    vector_group: VectorGroup | None = None
    neutral_connected: bool = False  # indicates if neutral line is connected to winding object



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from typing import TYPE_CHECKING

from loguru import logger

from psdm.base import Base
from psdm.base import UniqueTuple
from psdm.meta import Meta
from psdm.steadystate_case.external_grid import ExternalGrid
from psdm.steadystate_case.load import Load
from psdm.steadystate_case.transformer import Transformer

if TYPE_CHECKING:
    from psdm.topology.topology import Topology


class Case(Base):
    """This class represents a steady state case / an operation point of a grid.

    It is characterized by the operating points of all loads, transformers and external grids in operation.
    """

    meta: Meta
    loads: UniqueTuple[Load]
    transformers: UniqueTuple[Transformer]
    external_grids: UniqueTuple[ExternalGrid]

    def is_valid_topology(self, topology: Topology) -> bool:
        logger.info("Verifying steadystate case ...")
        if topology.meta != self.meta:
            logger.error("Metadata does not match.")
            return False

        if not self._is_proper_element_number(topology):
            return False

        if not self._is_proper_elements(topology):
            return False

        logger.info("Verifying steadystate case was successful.")
        return True

    def _is_proper_element_number(self, topology: Topology) -> bool:
        if len(self.loads) != len(topology.loads):
            logger.error(
                "Number of loads does not match. Is {n_act}, should be {n_ref}.",
                n_act=len(self.loads),
                n_ref=len(topology.loads),
            )
            return False

        if len(self.transformers) != len(topology.transformers):
            logger.error(
                "Number of transformers does not match. Is {n_act}, should be {n_ref}.",
                n_act=len(self.transformers),
                n_ref=len(topology.transformers),
            )
            return False

        if len(self.external_grids) != len(topology.external_grids):
            logger.error(
                "Number of external grids does not match. Is {n_act}, should be {n_ref}.",
                n_act=len(self.external_grids),
                n_ref=len(topology.external_grids),
            )
            return False

        return True

    def _is_proper_elements(self, topology: Topology) -> bool:
        if not self._is_proper_loads(topology):
            return False

        if not self._is_proper_transformers(topology):
            return False

        return self._is_proper_external_grids(topology)

    def _is_proper_loads(self, topology: Topology) -> bool:
        load_names = [e.name for e in self.loads]
        for load in topology.loads:
            if load.name not in load_names:
                logger.error("Load {load_name} is not in steadystate case.", load_name=load.name)
                return False

        return True

    def _is_proper_transformers(self, topology: Topology) -> bool:
        transformer_names = [e.name for e in self.transformers]
        for trafo in topology.transformers:
            if trafo.name not in transformer_names:
                logger.error("Transformer {trafo_name} is not in steadystate case.", trafo_name=trafo.name)
                return False

        return True

    def _is_proper_external_grids(self, topology: Topology) -> bool:
        external_grid_names = [e.name for e in self.external_grids]
        for ext_grid in topology.external_grids:
            if ext_grid.name not in external_grid_names:
                logger.error(
                    "External grid {ext_grid_name} is not in steadystate case.",
                    ext_grid_name=ext_grid.name,
                )
                return False

        return True



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from psdm.base import Base


class Node(Base):
    """This class represents a terminal within the grid."""

    name: str
    u_n: float
    description: str | None = None



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from psdm.base import Base
from psdm.topology.load_model import CONSTANT_POWER_LM
from psdm.topology.load_model import LoadModel


class ReactivePower(Base):
    """This class represents the reactive power characteristic of a load.

    If no load model is explicitly specified by the user, a constant power characteristic is assumed.
    """

    load_model: LoadModel = CONSTANT_POWER_LM



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

import enum
import json
import pathlib
import typing as t

import pydantic
from pydantic_core import PydanticCustomError

T = t.TypeVar("T")
U = t.TypeVar("U", bound=t.Hashable)


class VoltageSystemType(enum.Enum):
    AC = "AC"
    DC = "DC"


class CosphiDir(enum.Enum):
    UE = "UE"
    OE = "OE"


def _validate_unique_list(v: tuple[U]) -> tuple[U]:
    if len(v) != len(set(v)):
        error_type = "unique_list"
        message_template = "List must be unique"
        raise PydanticCustomError(error_type=error_type, message_template=message_template)
    return v


UniqueTuple = t.Annotated[
    tuple[U, ...],
    pydantic.AfterValidator(_validate_unique_list),
    pydantic.Field(json_schema_extra={"uniqueItems": True}),
]


class Base(pydantic.BaseModel):
    model_config = {
        "frozen": True,
        "use_enum_values": True,
    }

    @classmethod
    def from_file(cls, file_path: str | pathlib.Path) -> Base:
        file_path = pathlib.Path(file_path)
        with file_path.open("r", encoding="utf-8") as file_handle:
            return cls.model_validate_json(file_handle.read())

    def to_json(self, file_path: str | pathlib.Path, indent: int = 2) -> None:
        file_path = pathlib.Path(file_path)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with file_path.open("w+", encoding="utf-8") as file_handle:
            _json_data = self.model_dump_json()
            _data = json.loads(_json_data)
            json.dump(_data, file_handle, indent=indent, sort_keys=True)

    @classmethod
    def from_json(cls, json_str: str) -> Base:
        return cls.model_validate_json(json_str)



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from contextlib import nullcontext as does_not_raise

import pydantic
import pytest

from psdm.topology.load import RatedPower
from psdm.topology.load import PowerType


class TestRatedPower:
    @pytest.mark.parametrize(
        (
            "value",
            "value_a",
            "value_b",
            "value_c",
            "cosphi",
            "cosphi_a",
            "cosphi_b",
            "cosphi_c",
            "power_type",
            "is_symmetrical",
            "expectation",
        ),
        [
            (0, 0, 0, 0, 0, 0, 0, 0, PowerType.AC_APPARENT, True, does_not_raise()),
            (3, 1, 1, 1, 1, 1, 1, 1, PowerType.AC_APPARENT, True, does_not_raise()),
            (4, 2, 1, 1, 1, 1, 1, 1, PowerType.AC_APPARENT, False, does_not_raise()),
            (2, 2, 1, 1, 1, 1, 2, 1, PowerType.AC_APPARENT, False, pytest.raises(pydantic.ValidationError)),
            (2, -2, 1, 1, 1, 1, 1, 1, PowerType.AC_APPARENT, True, pytest.raises(pydantic.ValidationError)),
            (0, -2, 1, 1, 1, 1, 1, 1, PowerType.AC_APPARENT, False, pytest.raises(pydantic.ValidationError)),
            (4, 2, 1, 1, 1, 1, 2, 1, PowerType.AC_APPARENT, True, pytest.raises(pydantic.ValidationError)),
            (0, -2, 1, 1, 1, 1, 1, 1, PowerType.AC_APPARENT, True, pytest.raises(pydantic.ValidationError)),
            (3, 1, 1, 1, 1, 1, 1, 1, PowerType.AC_APPARENT, False, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(  # noqa: PLR0913
        self,
        value,
        value_a,
        value_b,
        value_c,
        cosphi,
        cosphi_a,
        cosphi_b,
        cosphi_c,
        power_type,
        is_symmetrical,
        expectation,
    ) -> None:
        with expectation:
            RatedPower(
                value=value,
                value_a=value_a,
                value_b=value_b,
                value_c=value_c,
                cosphi=cosphi,
                cosphi_a=cosphi_a,
                cosphi_b=cosphi_b,
                cosphi_c=cosphi_c,
                power_type=power_type,
                is_symmetrical=is_symmetrical,
            )



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from enum import Enum

from psdm.base import Base


class GridType(Enum):
    SL = "SL"  # slack node: voltage amplitude and phase angle is fixed
    PV = "PV"  # active power and voltage amplitude is fixed
    PQ = "PQ"  # active power and reactive power is fixed


class ExternalGrid(Base):
    """This class represents an external grid or a grid subsitute equivalent respectively.

    It is characterized by a grid type (slack, P-V-node, P-Q-node).
    """

    name: str
    description: str | None
    node: str
    type: GridType  # noqa: A003
    short_circuit_power_max: float
    short_circuit_power_min: float



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from psdm.base import Base


class Transformer(Base):
    """This class represents the operationg point of a transformer."""

    name: str
    tap_pos: int | None = None  # actual tap position



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

import enum
import typing as t

import pydantic

from psdm.base import Base
from psdm.base import VoltageSystemType
from psdm.topology.active_power import ActivePower
from psdm.topology.reactive_power import ReactivePower


class LoadType(enum.Enum):
    CONSUMER = "CONSUMER"
    PRODUCER = "PRODUCER"
    PROSUMER = "PROSUMER"
    SHUNT = "SHUNT"
    STORAGE = "STORAGE"


class SystemType(enum.Enum):
    BATTERY_STORAGE = "BATTERY_STORAGE"
    BIOGAS = "BIOGAS"
    CHARGING_POINT = "CHARGING_POINT"
    COAL = "COAL"
    DIESEL = "DIESEL"
    EXTERNAL_GRID_EQUIVALENT = "EXTERNAL_GRID_EQUIVALENT"
    FILTER_C = "FILTER_C"
    FILTER_HARMONIC = "FILTER_HARMONIC"
    FILTER_RL = "FILTER_RL"
    FILTER_RLC = "FILTER_RLC"
    FILTER_RLCCRP = "FILTER_RLCCRP"
    FILTER_RLCRP = "FILTER_RLCRP"
    FIXED_CONSUMPTION = "FIXED_CONSUMPTION"
    FUELCELL = "FUELCELL"
    GAS = "GAS"
    HEAT_PUMP = "HEAT_PUMP"
    HVAC = "HVAC"
    HVDC = "HVDC"
    HYDRO = "HYDRO"
    NIGHT_STORAGE = "NIGHT_STORAGE"
    NUCLEAR = "NUCLEAR"
    OIL = "OIL"
    OTHER = "OTHER"
    PEAT = "PEAT"
    PUMP_STORAGE = "PUMP_STORAGE"
    PV = "PV"
    REACTIVE_POWER_COMPENSATOR = "REACTIVE_POWER_COMPENSATOR"
    RENEWABLE_ENERGY = "RENEWABLE_ENERGY"
    SOLAR = "SOLAR"
    STAT_GEN = "STAT_GEN"
    STATIC_VAR_SYSTEM = "STATIC_VAR_SYSTEM"
    VARIABLE_CONSUMPTION = "VARIABLE_CONSUMPTION"
    WIND = "WIND"


class PhaseConnectionType(enum.Enum):
    ONE_PH_PH_E = "ONE_PH_PH_E"
    ONE_PH_PH_N = "ONE_PH_PH_N"
    ONE_PH_PH_PH = "ONE_PH_PH_PH"
    THREE_PH_D = "THREE_PH_D"
    THREE_PH_PH_E = "THREE_PH_PH_E"
    THREE_PH_YN = "THREE_PH_YN"
    TWO_PH_PH_E = "TWO_PH_PH_E"
    TWO_PH_YN = "TWO_PH_YN"


class Phase(enum.Enum):
    A = "A"
    B = "B"
    C = "C"
    N = "N"


class PowerType(enum.Enum):
    AC_ACTIVE = "AC_ACTIVE"
    AC_APPARENT = "AC_APPARENT"
    AC_REACTIVE = "AC_REACTIVE"
    CURRENT = "CURRENT"
    DC = "DC"
    GAS = "GAS"
    IMPEDANCE = "IMPEDANCE"
    MECHANICAL = "MECHANICAL"
    THERMAL = "THERMAL"


THRESHOLD = 0.51  # acceptable rounding error (0.5 W) + epsilon for calculation accuracy (0.01 W)


class PowerBase(Base):
    value: float  # power (three-phase)
    value_a: float  # power (phase a)
    value_b: float  # power (phase b)
    value_c: float  # power (phase c)
    is_symmetrical: bool


T = t.TypeVar("T", bound=PowerBase)


def validate_total(power: T) -> T:
    pow_total = power.value_a + power.value_b + power.value_c
    diff = abs(power.value - pow_total)
    if diff > THRESHOLD:
        msg = f"Power mismatch: Total power should be {pow_total}, is {power.value}."
        raise ValueError(msg)

    return power


def validate_symmetry(power: T) -> T:
    if power.value != 0:
        if power.is_symmetrical:
            if not (power.value_a == power.value_b == power.value_c):
                msg = "Power mismatch: Three-phase power of load is not symmetrical."
                raise ValueError(msg)

        elif power.value_a == power.value_b == power.value_c:
            msg = "Power mismatch: Three-phase power of load is symmetrical."
            raise ValueError(msg)

    return power


class RatedPower(PowerBase):
    value: float = pydantic.Field(..., ge=0)  # rated power (three phase); base for p.u. calculation
    value_a: float = pydantic.Field(..., ge=0)  # rated power (phase a)
    value_b: float = pydantic.Field(..., ge=0)  # rated power (phase b)
    value_c: float = pydantic.Field(..., ge=0)  # rated power (phase c)
    cosphi: float = pydantic.Field(1, ge=0, le=1)  # rated cos(phi) in relation to rated power
    cosphi_a: float = pydantic.Field(1, ge=0, le=1)  # rated cos(phi) (phase a)
    cosphi_b: float = pydantic.Field(1, ge=0, le=1)  # rated cos(phi) (phase b)
    cosphi_c: float = pydantic.Field(1, ge=0, le=1)  # rated cos(phi) (phase c)
    power_type: PowerType
    is_symmetrical: bool

    @pydantic.model_validator(mode="after")  # type: ignore[arg-type]
    def _validate_symmetry(cls, power: RatedPower) -> RatedPower:
        return validate_symmetry(power)

    @pydantic.model_validator(mode="after")  # type: ignore[arg-type]
    def _validate_total(cls, power: RatedPower) -> RatedPower:
        return validate_total(power)


class ConnectedPhases(Base):
    phases_a: tuple[Phase, Phase] | None
    phases_b: tuple[Phase, Phase] | None
    phases_c: tuple[Phase, Phase] | None

    @pydantic.computed_field  # type: ignore[misc]
    @property
    def n_connected_phases(self) -> int:
        return sum([getattr(self, f"phases_{idx}") is not None for idx in ["a", "b", "c"]])


class Load(Base):  # including assets of type load and generator
    """This class represents a load.

    It is mainly characterized by the load type (Producer, Consumer, Storage or passive shunt).
    """

    name: str
    node: str
    u_n: float  # nominal voltage of the connected node
    rated_power: RatedPower
    active_power: ActivePower
    reactive_power: ReactivePower
    type: LoadType  # noqa: A003
    connected_phases: ConnectedPhases
    system_type: SystemType
    phase_connection_type: PhaseConnectionType
    voltage_system_type: VoltageSystemType
    description: str | None = None



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :author: Sebastian Krahmer <sebastian.krahmer@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

import enum

import pydantic

from psdm.base import Base
from psdm.base import CosphiDir
from psdm.steadystate_case.characteristic import Characteristic
from psdm.topology.load import PowerBase
from psdm.topology.load import validate_symmetry
from psdm.topology.load import validate_total


class QControlStrategy(enum.Enum):
    COSPHI_CONST = "COSPHI_CONST"
    COSPHI_P = "COSPHI_P"
    COSPHI_U = "COSPHI_U"
    ND = "ND"
    Q_CONST = "Q_CONST"
    Q_P = "Q_P"
    Q_U = "Q_U"
    TANPHI_CONST = "TANPHI_CONST"
    U_CONST = "U_CONST"


class PControlStrategy(enum.Enum):
    ND = "ND"
    P_CONST = "P_CONST"
    P_F = "P_F"


class ControlledVoltageRef(enum.Enum):
    A = "A"
    AB = "AB"
    AVG = "AVG"
    B = "B"
    BC = "BC"
    C = "C"
    CA = "CA"
    POS_SEQ = "POS_SEQ"


class ControlQConst(Base):
    # q-setpoint control mode
    value: float  # set point of reactive power (three-phase)
    value_a: float  # set point of reactive power (phase a)
    value_b: float  # set point of reactive power (phase b)
    value_c: float  # pset point of reactive ower (phase c)
    is_symmetrical: bool

    control_strategy: QControlStrategy = QControlStrategy.Q_CONST

    @pydantic.model_validator(mode="after")  # type: ignore[arg-type]
    def _validate_symmetry(cls, controller: ControlQConst) -> PowerBase:
        power = PowerBase(
            value=controller.value,
            value_a=controller.value_a,
            value_b=controller.value_b,
            value_c=controller.value_c,
            is_symmetrical=controller.is_symmetrical,
        )
        return validate_symmetry(power)

    @pydantic.model_validator(mode="after")  # type: ignore[arg-type]
    def _validate_total(cls, controller: ControlQConst) -> PowerBase:
        power = PowerBase(
            value=controller.value,
            value_a=controller.value_a,
            value_b=controller.value_b,
            value_c=controller.value_c,
            is_symmetrical=controller.is_symmetrical,
        )
        return validate_total(power)


class ControlUConst(Base):
    # u-setpoint control mode
    u_set: float = pydantic.Field(ge=0)  # Setpoint of voltage.
    u_meas_ref: ControlledVoltageRef = ControlledVoltageRef.POS_SEQ  # voltage reference

    control_strategy: QControlStrategy = QControlStrategy.U_CONST


class ControlTanphiConst(Base):
    # cos(phi) control mode
    cosphi_dir: CosphiDir
    cosphi: float = pydantic.Field(ge=0, le=1)  # cos(phi) for calculation of Q in relation to P.

    control_strategy: QControlStrategy = QControlStrategy.TANPHI_CONST


class ControlCosphiConst(Base):
    # cos(phi) control mode
    cosphi_dir: CosphiDir
    cosphi: float = pydantic.Field(ge=0, le=1)  # cos(phi) for calculation of Q in relation to P.

    control_strategy: QControlStrategy = QControlStrategy.COSPHI_CONST


class ControlCosphiP(Base):
    # cos(phi(P)) control mode
    cosphi_ue: float = pydantic.Field(
        ge=0,
        le=1,
    )  # under excited: cos(phi) for calculation of Q in relation to P.
    cosphi_oe: float = pydantic.Field(
        ge=0,
        le=1,
    )  # over excited: cos(phi) for calculation of Q in relation to P.
    p_threshold_ue: float = pydantic.Field(le=0)  # under excited: threshold for P.
    p_threshold_oe: float = pydantic.Field(le=0)  # over excited: threshold for P.

    control_strategy: QControlStrategy = QControlStrategy.COSPHI_P


class ControlCosphiU(Base):
    # cos(phi(U)) control mode
    cosphi_ue: float = pydantic.Field(
        ...,
        ge=0,
        le=1,
    )  # under excited: cos(phi) for calculation of Q in relation to P.
    cosphi_oe: float = pydantic.Field(
        ...,
        ge=0,
        le=1,
    )  # over excited: cos(phi) for calculation of Q in relation to P.
    u_threshold_ue: float = pydantic.Field(..., ge=0)  # under excited: threshold for U.
    u_threshold_oe: float = pydantic.Field(..., ge=0)  # over excited: threshold for U.

    control_strategy: QControlStrategy = QControlStrategy.COSPHI_U


class ControlQU(Base):
    # Q(U) characteristic control mode
    droop_tg_2015: float = pydantic.Field(
        ...,
        ge=0,
    )  # Droop/Slope based on technical guideline VDE-AR-N 4120:2015: '%/kV'-value --> Q = m_% * Pr * dU_kV
    droop_tg_2018: float = pydantic.Field(
        ...,
        ge=0,
    )  # Droop/Slope based on technical guideline VDE-AR-N 4120:2018: '%/pu'-value --> Q = m_% * Pr * dU_(% of Un)
    u_q0: float = pydantic.Field(..., ge=0)  # Voltage value, where Q=0: absolut value in V
    u_deadband_up: float = pydantic.Field(
        ...,
        ge=0,
    )  # Width of upper deadband (U_1_up - U_Q0): absolut value in V
    u_deadband_low: float = pydantic.Field(
        ...,
        ge=0,
    )  # Width of lower deadband (U_Q0 - U_1_low): absolut value in V
    q_max_ue: float = pydantic.Field(..., ge=0)  # Under excited limit of Q: absolut value in var
    q_max_oe: float = pydantic.Field(..., ge=0)  # Over excited limit of Q: absolut value in var

    control_strategy: QControlStrategy = QControlStrategy.Q_U


def validate_pos(value: float | None) -> float | None:
    if value is not None and value < 0:
        raise ValueError

    return value


class ControlQP(Base):
    # Q(P) characteristic control mode
    q_p_characteristic: Characteristic
    q_max_ue: float | None = None  # Under excited limit of Q: absolut value
    q_max_oe: float | None = None  # Over excited limit of Q: absolut value

    control_strategy: QControlStrategy = QControlStrategy.Q_P

    @pydantic.field_validator("q_max_ue", mode="before")
    def validate_q_max_ue(cls, v: float | None) -> float | None:
        return validate_pos(v)

    @pydantic.field_validator("q_max_oe", mode="before")
    def validate_q_max_oe(cls, v: float | None) -> float | None:
        return validate_pos(v)


class ControlPConst(Base):
    # p-setpoint control mode
    value: float  # set point of active power (three-phase)
    value_a: float  # set point of active power (phase a)
    value_b: float  # set point of active power (phase b)
    value_c: float  # pset point of active ower (phase c)
    is_symmetrical: bool

    control_strategy: PControlStrategy = PControlStrategy.P_CONST

    @pydantic.model_validator(mode="after")  # type: ignore[arg-type]
    def _validate_symmetry(cls, controller: ControlPConst) -> PowerBase:
        power = PowerBase(
            value=controller.value,
            value_a=controller.value_a,
            value_b=controller.value_b,
            value_c=controller.value_c,
            is_symmetrical=controller.is_symmetrical,
        )
        return validate_symmetry(power)

    @pydantic.model_validator(mode="after")  # type: ignore[arg-type]
    def _validate_total(cls, controller: ControlPConst) -> PowerBase:
        power = PowerBase(
            value=controller.value,
            value_a=controller.value_a,
            value_b=controller.value_b,
            value_c=controller.value_c,
            is_symmetrical=controller.is_symmetrical,
        )
        return validate_total(power)


class ControlPF(Base):
    # P(f) characteristic control mode
    droop_over_freq: float = pydantic.Field(
        ...,
        ge=0,
    )  # Droop/Slope of power infeed reduction if freqeuncy is above f_deadband_up: '%/Hz'
    droop_under_freq: float = pydantic.Field(
        ...,
        ge=0,
    )  # Droop/Slope of power infeed increase if freqeuncy is below f_deadband_low: '%/Hz'
    f_p0: float = pydantic.Field(..., ge=0)  # Nominal freqeuncy value: absolut value in Hz
    f_deadband_up: float = pydantic.Field(
        ...,
        ge=0,
    )  # Width of upper deadband (f_up - f_P0): absolut value in Hz
    f_deadband_low: float = pydantic.Field(
        ...,
        ge=0,
    )  # Width of lower deadband (f_P0 - f_low): absolut value in Hz

    control_strategy: PControlStrategy = PControlStrategy.P_F


QControlType = (
    ControlQConst
    | ControlUConst
    | ControlTanphiConst
    | ControlCosphiConst
    | ControlCosphiP
    | ControlCosphiU
    | ControlQU
    | ControlQP
)

PControlType = ControlPConst | ControlPF


class Controller(Base):
    node_target: str  # the controlled node (which can be differ from node the load is connected to)
    external_controller_name: str | None = None  # if external controller is specified --> name


class PController(Controller):
    """This class represents a controller of active power of a load.

    It is characterized by the control type, which comes with different controller parameters.
    """

    control_type: PControlType


class QController(Controller):
    """This class represents a controller of reactive power of a load.

    It is characterized by the control type, which comes with different controller parameters.
    """

    control_type: QControlType



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from psdm.base import VoltageSystemType
from psdm.topology.branch import Branch, BranchType


class TestBranch:
    def test_init(self) -> None:
        Branch(
            node_1="asd",
            node_2="fgh",
            name="wqertasd",
            u_n=1,
            i_r=1,
            b1=1,
            g1=1,
            x1=1,
            r1=1,
            type=BranchType.LINE,
            voltage_system_type=VoltageSystemType.AC,
        )



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from psdm.base import Base


class ExternalGrid(Base):
    """This class represents the operating point of an external grid or a grid subsitute equivalent respectively."""

    name: str
    u_0: float | None = None
    phi_0: float | None = None
    p_0: float | None = None
    q_0: float | None = None



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from psdm.base import Base
from psdm.steadystate_case.active_power import ActivePower
from psdm.steadystate_case.reactive_power import ReactivePower


class Load(Base):  # including assets of type load and generator
    """This class represents the operating point of a load.

    It is characterized by the active and reactive power.
    """

    name: str
    active_power: ActivePower
    reactive_power: ReactivePower



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from psdm.base import Base


class Switch(Base):
    """This class represents an imaginary switching element."""

    element: str
    node: str
    state: bool  # 0:opened; 1:closed



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

from enum import Enum

from psdm.base import Base
from psdm.base import VoltageSystemType


class BranchType(Enum):
    LINE = "LINE"
    COUPLER = "COUPLER"
    FUSE = "FUSE"


class Branch(Base):
    """This class represents a branch adn therefore includes lines, cables or branch fuses.

    It is characterized by a branch type (line, cable or fuse).
    """

    node_1: str
    node_2: str
    name: str
    u_n: float  # nominal voltage of the branch connected nodes
    i_r: float | None  # rated current of branch (thermal limit in continuous operation)
    r1: float  # positive sequence values of PI-representation
    x1: float  # positive sequence values of PI-representation
    g1: float  # positive sequence values of PI-representation
    b1: float  # positive sequence values of PI-representation
    type: BranchType  # noqa: A003
    voltage_system_type: VoltageSystemType
    r0: float | None = None  # zero sequence values of PI-representation
    x0: float | None = None  # zero sequence values of PI-representation
    g0: float | None = None  # zero sequence values of PI-representation
    b0: float | None = None  # zero sequence values of PI-representation
    f_n: float | None = None  # nominal frequency the values x and b apply
    description: str | None = None
    energized: bool | None = None
    length: float | None = None  # length of the line the impedance and admittance values apply
    rn: float | None = None  # neutral natural values
    xn: float | None = None  # neutral natural values
    gn: float | None = None  # neutral natural values
    bn: float | None = None  # neutral natural values
    rpn: float | None = None  # neutral-line couple values
    xpn: float | None = None  # neutral-line couple values
    gpn: float | None = None  # neutral-line couple values
    bpn: float | None = None  # neutral-line couple values



# :author: Sasan Jacob Rasti <sasan_jacob.rasti@tu-dresden.de>
# :author: Sebastian Krahmer <sebastian.krahmer@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from __future__ import annotations

import enum

from psdm.base import Base
from psdm.base import UniqueTuple
from psdm.topology.windings import Winding


class TapSide(enum.Enum):
    HV = "HV"
    MV = "MV"
    LV = "LV"


class TransformerPhaseTechnologyType(enum.Enum):
    SINGLE_PH_E = "SINGLE_PH_E"
    SINGLE_PH = "SINGLE_PH"
    THREE_PH = "THREE_PH"


class VectorGroup(enum.Enum):
    Dd0 = "Dd0"
    Yy0 = "Yy0"
    YNy0 = "YNy0"
    Yyn0 = "Yyn0"
    YNyn0 = "YNyn0"
    Dz0 = "Dz0"
    Dzn0 = "Dzn0"
    Zd0 = "Zd0"
    ZNd0 = "ZNd0"
    Dyn1 = "Dyn1"
    Dy5 = "Dy5"
    Dyn5 = "Dyn5"
    Yd5 = "Yd5"
    YNd5 = "YNd5"
    Yz5 = "Yz5"
    YNz5 = "YNz5"
    Yzn5 = "Yzn5"
    YNzn5 = "YNzn5"
    Dd6 = "Dd6"
    Yy6 = "Yy6"
    YNy6 = "YNy6"
    Yyn6 = "Yyn6"
    YNyn6 = "YNyn6"
    Dz6 = "Dz6"
    Dzn6 = "Dzn6"
    Zd6 = "Zd6"
    ZNd6 = "ZNd6"
    Dyn7 = "Dyn7"
    Dy11 = "Dy11"
    Dyn11 = "Dyn11"
    Yd11 = "Yd11"
    YNd11 = "YNd11"
    Yz11 = "Yz11"
    YNz11 = "YNz11"
    Yzn11 = "Yzn11"
    YNzn11 = "YNzn11"


class Transformer(Base):
    """This class represents a transformer and consists of winding elements.

    It is characterized by windings elements (2w or 3w), the vector group as well as the transformer tap control.
    """

    node_1: str
    node_2: str
    name: str
    number: int  # number of parallel units
    vector_group: VectorGroup  # specifier for connection of wiring e.g. DYn5
    i_0: float  # no-load current in %
    p_fe: float  # no-load losses (Iron losses)
    i_00: float | None = None  # zero sequence values of PI-representation
    p_fe0: float | None = None  # zero sequence values of PI-representation
    windings: UniqueTuple[Winding]  # winding object for each voltage level
    phase_technology_type: TransformerPhaseTechnologyType | None = None  # three- or single-phase-transformer
    description: str | None = None
    tap_u_abs: float | None = None  # voltage deviation per tap position change in %
    tap_u_phi: float | None = None  # voltage angle deviation per tap position in °
    tap_max: int | None = None  # upper position of tap for tap control
    tap_min: int | None = None  # lower position of tap for tap control
    tap_neutral: int | None = None  # initial position where rated transformation ratio is specified
    tap_side: TapSide | None = None  # transformer side of where tap changer is installed



# :author: Sebastian Krahmer <sebastian.krahmer@tu-dresden.de>
# :copyright: Copyright (c) Institute of Electrical Power Systems and High Voltage Engineering - TU Dresden, 2022-2023.
# :license: BSD 3-Clause

from contextlib import nullcontext as does_not_raise

import pydantic
import pytest

from psdm.base import CosphiDir
from psdm.steadystate_case.characteristic import Characteristic
from psdm.steadystate_case.controller import (
    ControlCosphiConst,
    ControlCosphiP,
    ControlCosphiU,
    ControlPConst,
    ControlPF,
    ControlledVoltageRef,
    ControlQConst,
    ControlQP,
    ControlQU,
    ControlTanphiConst,
    ControlUConst,
)


class TestControlQConst:
    @pytest.mark.parametrize(
        (
            "value",
            "value_a",
            "value_b",
            "value_c",
            "is_symmetrical",
            "expectation",
        ),
        [
            (-1500, -500, -500, -500, True, does_not_raise()),
            (-1500, -500, -500, -500, False, pytest.raises(pydantic.ValidationError)),
            (1500, 500, 500, 500, True, does_not_raise()),
            (1500, 600, 500, 500, True, pytest.raises(pydantic.ValidationError)),
            (1500, 500, 1000, 0, False, does_not_raise()),
            (1700, 500, 1000, 200, True, pytest.raises(pydantic.ValidationError)),
            (1700, 500, 1000, 100, False, pytest.raises(pydantic.ValidationError)),
            (0, 0, 0, 0, True, does_not_raise()),
            (None, None, None, None, None, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(
        self,
        value,
        value_a,
        value_b,
        value_c,
        is_symmetrical,
        expectation,
    ) -> None:
        with expectation:
            ControlQConst(
                node_target="A",
                value=value,
                value_a=value_a,
                value_b=value_b,
                value_c=value_c,
                is_symmetrical=is_symmetrical,
            )


class TestControlUConst:
    @pytest.mark.parametrize(
        (
            "u_set",
            "u_meas_ref",
            "expectation",
        ),
        [
            (110000, ControlledVoltageRef.POS_SEQ, does_not_raise()),
            (100000, ControlledVoltageRef.AVG, does_not_raise()),
            (110000, None, pytest.raises(pydantic.ValidationError)),
            (
                -110000,
                ControlledVoltageRef.POS_SEQ,
                pytest.raises(pydantic.ValidationError),
            ),
            (None, ControlledVoltageRef.POS_SEQ, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(
        self,
        u_set,
        u_meas_ref,
        expectation,
    ) -> None:
        with expectation:
            ControlUConst(
                node_target="A",
                u_set=u_set,
                u_meas_ref=u_meas_ref,
            )


class TestControlTanphiConst:
    @pytest.mark.parametrize(
        (
            "cosphi_dir",
            "cosphi",
            "expectation",
        ),
        [
            (CosphiDir.UE, 1, does_not_raise()),
            (CosphiDir.UE, 0, does_not_raise()),
            (CosphiDir.OE, 0.9, does_not_raise()),
            (None, 0.9, pytest.raises(pydantic.ValidationError)),
            (CosphiDir.OE, -0.9, pytest.raises(pydantic.ValidationError)),
            (CosphiDir.OE, 2, pytest.raises(pydantic.ValidationError)),
            (CosphiDir.OE, None, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(
        self,
        cosphi_dir,
        cosphi,
        expectation,
    ) -> None:
        with expectation:
            ControlTanphiConst(
                node_target="A",
                cosphi_dir=cosphi_dir,
                cosphi=cosphi,
            )


class TestControlCosphiConst:
    @pytest.mark.parametrize(
        (
            "cosphi_dir",
            "cosphi",
            "expectation",
        ),
        [
            (CosphiDir.UE, 1, does_not_raise()),
            (CosphiDir.UE, 0, does_not_raise()),
            (CosphiDir.OE, 0.9, does_not_raise()),
            (None, 0.9, pytest.raises(pydantic.ValidationError)),
            (CosphiDir.OE, -0.9, pytest.raises(pydantic.ValidationError)),
            (CosphiDir.OE, 2, pytest.raises(pydantic.ValidationError)),
            (CosphiDir.OE, None, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(
        self,
        cosphi_dir,
        cosphi,
        expectation,
    ) -> None:
        with expectation:
            ControlCosphiConst(
                node_target="A",
                cosphi_dir=cosphi_dir,
                cosphi=cosphi,
            )


class TestControlCosphiP:
    @pytest.mark.parametrize(
        (
            "cosphi_ue",
            "cosphi_oe",
            "p_threshold_ue",
            "p_threshold_oe",
            "expectation",
        ),
        [
            (0.9, 0.9, -1, -2, does_not_raise()),
            (1, 0, -2, -1, does_not_raise()),
            (0.9, 1.1, -2, -1, pytest.raises(pydantic.ValidationError)),
            (1.1, 0.9, -2, -1, pytest.raises(pydantic.ValidationError)),
            (-0.9, 0.9, -2, -1, pytest.raises(pydantic.ValidationError)),
            (0.9, -0.9, -2, -1, pytest.raises(pydantic.ValidationError)),
            (0.9, 0.9, 2, -1, pytest.raises(pydantic.ValidationError)),
            (0.9, 0.9, -2, 1, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(  # noqa: PLR0913
        self,
        cosphi_ue,
        cosphi_oe,
        p_threshold_ue,
        p_threshold_oe,
        expectation,
    ) -> None:
        with expectation:
            ControlCosphiP(
                node_target="A",
                cosphi_ue=cosphi_ue,
                cosphi_oe=cosphi_oe,
                p_threshold_ue=p_threshold_ue,
                p_threshold_oe=p_threshold_oe,
            )


class TestControlCosphiU:
    @pytest.mark.parametrize(
        (
            "cosphi_ue",
            "cosphi_oe",
            "u_threshold_ue",
            "u_threshold_oe",
            "expectation",
        ),
        [
            (0.9, 0.9, 20000, 24000, does_not_raise()),
            (1, 0, 24000, 20000, does_not_raise()),
            (0.9, 1.1, 24000, 20000, pytest.raises(pydantic.ValidationError)),
            (1.1, 0.9, 24000, 20000, pytest.raises(pydantic.ValidationError)),
            (-0.9, 0.9, 24000, 20000, pytest.raises(pydantic.ValidationError)),
            (0.9, -0.9, 24000, 20000, pytest.raises(pydantic.ValidationError)),
            (0.9, 0.9, -24000, 20000, pytest.raises(pydantic.ValidationError)),
            (0.9, 0.9, 20000, -20000, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(  # noqa: PLR0913
        self,
        cosphi_ue,
        cosphi_oe,
        u_threshold_ue,
        u_threshold_oe,
        expectation,
    ) -> None:
        with expectation:
            ControlCosphiU(
                node_target="A",
                cosphi_ue=cosphi_ue,
                cosphi_oe=cosphi_oe,
                u_threshold_ue=u_threshold_ue,
                u_threshold_oe=u_threshold_oe,
            )


class TestControlQU:
    @pytest.mark.parametrize(
        (
            "droop_tg_2015",
            "droop_tg_2018",
            "u_q0",
            "u_deadband_up",
            "u_deadband_low",
            "q_max_ue",
            "q_max_oe",
            "expectation",
        ),
        [
            (5, 6, 110000, 1000, 1000, 10000, 10000, does_not_raise()),
            (5, 6, 110000, 1000, 2000, 20000, 10000, does_not_raise()),
            (-5, 6, 110000, 1000, 2000, 20000, 10000, pytest.raises(pydantic.ValidationError)),
            (5, -6, 110000, 1000, 2000, 20000, 10000, pytest.raises(pydantic.ValidationError)),
            (5, 6, -110000, 1000, 2000, 20000, 10000, pytest.raises(pydantic.ValidationError)),
            (5, 6, 110000, -1000, 2000, 20000, 10000, pytest.raises(pydantic.ValidationError)),
            (5, 6, 110000, 1000, -2000, 20000, 10000, pytest.raises(pydantic.ValidationError)),
            (5, 6, 110000, 1000, 2000, -20000, 10000, pytest.raises(pydantic.ValidationError)),
            (5, 6, 110000, 1000, 2000, 20000, -10000, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(  # noqa: PLR0913
        self,
        droop_tg_2015,
        droop_tg_2018,
        u_q0,
        u_deadband_up,
        u_deadband_low,
        q_max_ue,
        q_max_oe,
        expectation,
    ) -> None:
        with expectation:
            ControlQU(
                node_target="A",
                droop_tg_2015=droop_tg_2015,
                droop_tg_2018=droop_tg_2018,
                u_q0=u_q0,
                u_deadband_up=u_deadband_up,
                u_deadband_low=u_deadband_low,
                q_max_ue=q_max_ue,
                q_max_oe=q_max_oe,
            )


class TestControlQP:
    @pytest.mark.parametrize(
        (
            "q_p_characteristic",
            "q_max_ue",
            "q_max_oe",
            "expectation",
        ),
        [
            (Characteristic(name="Q_P_Char"), 10000, 10000, does_not_raise()),
            (Characteristic(name="Q_P_Char"), 10000, 20000, does_not_raise()),
            (None, 10000, 10000, pytest.raises(pydantic.ValidationError)),
            (Characteristic(name="Q_P_Char"), -10000, 10000, pytest.raises(pydantic.ValidationError)),
            (Characteristic(name="Q_P_Char"), 10000, -10000, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(  # noqa: PLR0913
        self,
        q_p_characteristic,
        q_max_ue,
        q_max_oe,
        expectation,
    ) -> None:
        with expectation:
            ControlQP(
                node_target="A",
                q_p_characteristic=q_p_characteristic,
                q_max_ue=q_max_ue,
                q_max_oe=q_max_oe,
            )


class TestControlPF:
    @pytest.mark.parametrize(
        (
            "droop_over_freq",
            "droop_under_freq",
            "f_p0",
            "f_deadband_up",
            "f_deadband_low",
            "expectation",
        ),
        [
            (40, 40, 50, 0.2, 0.2, does_not_raise()),
            (-40, 40, 50, 0.2, 0.2, pytest.raises(pydantic.ValidationError)),
            (40, -40, 50, 0.2, 0.2, pytest.raises(pydantic.ValidationError)),
            (40, 40, -50, 0.2, 0.2, pytest.raises(pydantic.ValidationError)),
            (40, 40, 50, -0.2, 0.2, pytest.raises(pydantic.ValidationError)),
            (40, 40, 50, 0.2, -0.2, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(  # noqa: PLR0913
        self,
        droop_over_freq,
        droop_under_freq,
        f_p0,
        f_deadband_up,
        f_deadband_low,
        expectation,
    ) -> None:
        with expectation:
            ControlPF(
                node_target="A",
                droop_over_freq=droop_over_freq,
                droop_under_freq=droop_under_freq,
                f_p0=f_p0,
                f_deadband_up=f_deadband_up,
                f_deadband_low=f_deadband_low,
            )


class TestControlPConst:
    @pytest.mark.parametrize(
        (
            "value",
            "value_a",
            "value_b",
            "value_c",
            "is_symmetrical",
            "expectation",
        ),
        [
            (-1500, -500, -500, -500, True, does_not_raise()),
            (-1500, -500, -500, -500, False, pytest.raises(pydantic.ValidationError)),
            (1500, 500, 500, 500, True, does_not_raise()),
            (1500, 600, 500, 500, True, pytest.raises(pydantic.ValidationError)),
            (1500, 500, 1000, 0, False, does_not_raise()),
            (1700, 500, 1000, 200, True, pytest.raises(pydantic.ValidationError)),
            (1700, 500, 1000, 100, False, pytest.raises(pydantic.ValidationError)),
            (0, 0, 0, 0, True, does_not_raise()),
            (None, None, None, None, None, pytest.raises(pydantic.ValidationError)),
        ],
    )
    def test_init(
        self,
        value,
        value_a,
        value_b,
        value_c,
        is_symmetrical,
        expectation,
    ) -> None:
        with expectation:
            ControlPConst(
                node_target="A",
                value=value,
                value_a=value_a,
                value_b=value_b,
                value_c=value_c,
                is_symmetrical=is_symmetrical,
            )
